# -*- coding: utf-8 -*-

# Copyright 2008-2010 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import cgi
import textwrap

class SkoolParser:
    DIRECTIVES = 'bcgituwz'

    def __init__(self, skoolfile, game_id, mode_args=(), tap_info=None, debug=False):
        self.game_id = game_id

        # mode_args may contain up to four items:
        #
        # mode:
        #   0 = do not parse asm directives
        #   1 = parse asm directives in @isub mode
        #   2 = parse asm directives in @ssub mode
        #   3 = parse asm directives in @rsub mode
        # verbose:
        #   whether to display info about instruction substitutions
        # do_warn:
        #   whether to show warnings
        # fix_mode:
        #   0 = apply no fixes
        #   1 = honour @ofix directives
        #   2 = honour @ofix and @bfix directives
        #   3 = honour @ofix, @bfix and @rfix directives
        self.mode = Mode(*mode_args)

        self.tap_info = tap_info
        self.debug = debug

        self.snapshot = [0] * 65536  # 64K of Spectrum memory
        self.instructions = {}       # address -> [Instructions]
        self.entries = {}            # address -> Routine/Data
        self.memory_map = []         # Routines and Datas
        self.tap_addresses = {}      # number -> command list address
        self.tap_descs = {}          # number -> command list description
        self.header = []
        self.stack = []
        self._parse_skool(skoolfile)

    def get_word(self, address):
        return self.snapshot[address] + 256 * self.snapshot[address + 1]

    def get_entry(self, address):
        return self.entries.get(str(address))

    def get_container(self, address):
        instructions = self.instructions.get(str(address))
        if instructions:
            return instructions[0].container

    def get_entry_point_refs(self, address):
        return [referrer.address for referrer in self.instructions[str(address)][0].referrers]

    def get_tap_refs(self, address):
        return self.entries[str(address)].tap_referrers

    def _parse_skool(self, skoolfile):
        map_entry = None
        instruction = None
        comments = []
        address_comments = []
        f = open(skoolfile, 'r')
        for line in f:
            if line[0] == ';':
                if self._parse_comment(line, comments):
                    instruction = None
                continue

            if not self.mode.include:
                continue

            if self.debug:
                sys.stderr.write(line)

            s_line = line.strip()
            if not s_line:
                instruction = None
                if comments:
                    if map_entry:
                        map_entry.end_comment = self._escape_comments(comments)
                    else:
                        self.header += comments
                comments[:] = []
                map_entry = None
                continue

            if s_line[0] == ';' and instruction:
                address_comments[-1][1] = '%s %s' % (address_comments[-1][1], s_line[1:].lstrip())
                continue

            # This line contains an instruction
            instruction, address_comment = self._parse_instruction(line)
            address = instruction.address
            ctl = instruction.ctl
            if ctl in SkoolParser.DIRECTIVES:
                desc, details, registers = self._parse_comment_block(comments)
                if ctl == 'c':
                    map_entry = Routine(address, ctl, desc, details, registers)
                else:
                    map_entry = Data(address, ctl, desc, details)
                self.entries[address] = map_entry
                self.memory_map.append(map_entry)
                comments[:] = []

            if map_entry:
                address_comments.append([instruction, address_comment])
                if not self.instructions.has_key(address):
                    self.instructions[address] = []
                self.instructions[address].append(instruction)
                map_entry.add_instruction(instruction)
                if comments:
                    comment = self._escape_comments(comments)
                    map_entry.add_mid_routine_comment(instruction.label, comment)
                    comments[:] = []

            # Set bytes in the snapshot if the instruction is DEF{B,M,S,W}
            if self.mode.html:
                operation = instruction.operation
                if operation[:4] in ('DEFB', 'DEFM', 'DEFS', 'DEFW'):
                    self._set_bytes(int(address), operation)
        f.close()

        # Do some post-processing
        self._parse_address_comments(address_comments)
        if self.mode.html:
            self._calculate_entry_sizes()
            self._calculate_references()
            if self.tap_info:
                self._calculate_tap_routine_usage()
        else:
            self._substitute_labels()

    def _escape_comments(self, comments):
        text = " ".join([comment.strip() for comment in comments])
        if self.mode.html:
            return cgi.escape(text)
        return text

    def _set_bytes(self, address, operation):
        if operation.startswith('DEFM'):
            start = operation.index('"') + 1
            message = operation[start:operation.index('"', start)]
            self.snapshot[address:address + len(message)] = [ord(c) for c in message]
            return
        args = [int(v) for v in operation[5:].split(',')]
        if operation.startswith('DEFS'):
            if len(args) > 1:
                span = args[0]
                self.snapshot[address:address + span] = [args[1]] * span
        else:
            words = operation[:4] == 'DEFW'
            for arg in args:
                self.snapshot[address] = arg % 256
                address += 1
                if words:
                    self.snapshot[address] = arg // 256
                    address += 1

    def _parse_comment(self, line, comments):
        """Parses a comment line. Returns False if the line contains an @ or !
        directive, and True otherwise."""
        comment = line[2:].rstrip()
        if comment.startswith('@'):
            self._parse_asm_directive(comment[1:])
            return False
        elif self.mode.started:
            if comment.startswith('!'):
                self._parse_skool_directive(comment[1:])
                return False
            elif self.mode.include:
                if self.debug:
                    sys.stderr.write(line)
                comments.append(comment)
        return True

    def _parse_skool_directive(self, directive):
        if directive.endswith('+end'):
            self.mode.include = True
        elif directive.endswith('+begin'):
            self.mode.include = directive.startswith(self.game_id)
        elif directive.startswith('refs='):
            for address in directive[5:].split(','):
                self.entries[address] = SkoolEntry(address)
        elif directive.startswith('defb='):
            address = int(directive[5:10])
            bytes = [int(b) for b in directive[11:].split(',')]
            self.snapshot[address:address + len(bytes)] = bytes

    def _parse_asm_directive(self, directive):
        if self.mode.started:
            prefix = directive[:4]
            infix = directive[len(prefix):len(prefix) + 1]
            suffix = directive[len(prefix) + len(infix):].rstrip()
            if prefix in ('ofix', 'bfix', 'rfix', 'isub', 'rsub') and infix in '+-' and suffix in ('begin', 'else', 'end'):
                cur_op = self.stack[-1] if self.stack else (None, None)
                if suffix == 'begin':
                    if prefix == cur_op[0]:
                        self._error('%s inside %s%s block' % (directive, cur_op[0], cur_op[1]))
                    self.stack.append((prefix, infix))
                elif suffix == 'else':
                    if cur_op[0] is None:
                        self._error('%s not inside block' % directive)
                    if prefix != cur_op[0] or infix == cur_op[1]:
                        self._error('%s inside %s%s block' % (directive, cur_op[0], cur_op[1]))
                    self.stack[-1] = (prefix, infix)
                elif suffix == 'end':
                    if cur_op[0] is None:
                        self._error('%s has no matching start directive' % directive)
                    if (prefix, infix) != cur_op:
                        self._error('%s cannot end %s block' % (directive, cur_op))
                    self.stack.pop()

                include = True
                for p, i in self.stack:
                    if p == 'ofix':
                        do_op = self.mode.do_ofixes
                    elif p == 'bfix':
                        do_op = self.mode.do_bfixes
                    elif p == 'rfix':
                        do_op = self.mode.do_rfixes
                    elif p == 'isub':
                        do_op = self.mode.asm
                    elif p == 'rsub':
                        do_op = self.mode.do_rsubs
                    include = i == '+' if do_op else i == '-'
                    if not include:
                        break;

                self.mode.include = include
                return

            if self.mode.asm and self.mode.include:
                if directive.startswith('rsub='):
                    self.mode.rsub = directive[5:].rstrip()
                elif directive.startswith('ssub='):
                    self.mode.ssub = directive[5:].rstrip()
                elif directive.startswith('isub='):
                    self.mode.isub = directive[5:].rstrip()
                elif directive.startswith('bfix='):
                    self.mode.bfix = directive[5:].rstrip()
                elif directive.startswith('ofix='):
                    self.mode.ofix = directive[5:].rstrip()
                elif directive.startswith('label='):
                    self.mode.label = directive[6:].rstrip()
                elif directive.startswith('rlabel='):
                    self.mode.rlabel = directive[7:].rstrip()
                elif directive.startswith('nolabel'):
                    self.mode.nolabel = True
                elif directive.startswith('nowarn'):
                    self.mode.nowarn = True
                elif directive.startswith('ignoreua'):
                    self.mode.ignoreua = True
                elif directive.startswith('ignoremrcua'):
                    self.mode.ignoremrcua = True
                elif directive.startswith('keep'):
                    self.mode.keep = True
                elif directive.startswith('org='):
                    self.mode.org = directive[4:].rstrip()
        elif directive.startswith('start'):
            self.mode.start()

    def _parse_instruction(self, line):
        ctl = line[0]
        address = line[1:6]
        comment_index = line.find(';')
        if comment_index < 0:
            comment_index = len(line)
        operation = line[7:comment_index].strip()
        address_comment = line[comment_index + 1:].strip()
        instruction = Instruction(ctl, address, operation)
        self.mode.apply_asm_attributes(instruction)
        if instruction.sub:
            self._info('%s %s -> %s' % (address, operation, instruction.sub))
        return instruction, address_comment

    def _parse_comment_block(self, comments):
        all = [[], [], []]
        index = 0
        last_line = ""
        for line in comments:
            if len(line) == 0:
                if len(last_line) > 0:
                    index += 1
                continue
            all[index].append(line)
            last_line = line
        registers = [reg_line.split(None, 1) for reg_line in all[2]]
        return [self._escape_comments(all[0]), self._escape_comments(all[1]), registers]

    def _parse_address_comments(self, comments):
        i = 0
        while i < len(comments):
            instruction, comment = comments[i]
            comment_lines = []
            if comment and comment.strip()[0] == '{':
                nesting = comment.count('{') - comment.count('}')
                while nesting > 0:
                    comment_lines.append(comments[i][1])
                    i += 1
                    nesting += comments[i][1].count('{') - comments[i][1].count('}')
                comment_lines.append(comments[i][1].strip('}'))
                comment_lines[0] = comment_lines[0].strip('{')
            else:
                comment_lines.append(comment)
            rowspan = len(comment_lines)
            address_comment = self._escape_comments(comment_lines).strip()
            instruction.set_comment(rowspan, address_comment)
            i += 1

    def _get_address(self, operation):
        address = ''
        started = False
        for c in operation:
            if c.isdigit():
                started = True
                address += c
            elif started:
                break
        return address

    def _calculate_entry_sizes(self):
        for i, entry in enumerate(self.memory_map[:-1]):
            prev_entry = self.memory_map[i - 1]
            prev_entry.size = int(entry.address) - int(prev_entry.address)
        last_entry = self.memory_map[-1]
        last_entry.size = 65536 - int(last_entry.address)

    def _calculate_references(self):
        # Parse operations for routine/data addresses
        for entry in self.memory_map:
            for instruction in entry.instructions:
                operation = instruction.operation
                if operation.startswith('DEFW') or operation[0:2] in ['JP', 'JR', 'CA']:
                    address = self._get_address(operation)
                    if not address:
                        continue
                    other_instructions = self.instructions.get(address, [])
                    if other_instructions:
                        other_entry = other_instructions[0].container
                    elif address in self.entries:
                        other_entry = self.entries[address]
                    else:
                        continue
                    if entry != other_entry:
                        instruction.set_reference(other_entry.address, address)
                        for other_instruction in other_instructions:
                            other_instruction.add_referrer(entry)

    def _info(self, s):
        if self.mode.verbose:
            sys.stderr.write('%s\n' % s)

    def _warn(self, s):
        if self.mode.warn:
            sys.stderr.write('WARNING: %s\n' % s)

    def _error(self, s):
        sys.stderr.write('ERROR: %s\n' % s)
        sys.exit(1)

    def _substitute_labels(self):
        self._generate_labels()
        for entry in self.memory_map:
            for instruction in entry.instructions:
                if not instruction.sub and instruction.keep:
                    self._info('Keeping operand: %s %s' % (instruction.address, instruction.operation))
                else:
                    self._label_operand(instruction)

    def _generate_labels(self):
        """Generate labels for mid-routine entry points (based on the label of
           the main entry point)."""
        for entry in self.entries.values():
            instructions = entry.instructions
            main_label = instructions[0].asm_label
            if main_label:
                index = 0
                for instruction in instructions[1:]:
                    if instruction.ctl in '!*' and not instruction.asm_label and not instruction.nolabel:
                        instruction.asm_label = '%s_%i' % (main_label, index)
                        index += 1

    def _label_operand(self, instruction):
        label_warn = instruction.sub is None and instruction.warn
        operation = instruction.operation
        operand = self._get_address(operation)
        instructions = self.instructions.get(operand)
        if instructions:
            reference = instructions[0]
            if reference.asm_label:
                rep = operation.replace(operand, reference.asm_label)
                if reference.is_in_routine() and label_warn and operation.startswith('LD '):
                    # Warn if a LD operand is replaced with a routine label in
                    # an unsubbed operation (will need @keep to retain operand,
                    # or @nowarn if the replacement is OK)
                    self._warn('LD operand replaced with routine label in unsubbed operation:\n  %s %s -> %s' % (instruction.address, operation, rep))
                instruction.operation = rep
            elif instruction.warn and instruction.is_in_routine():
                # Warn if we cannot find a label to replace the operand of this
                # routine instruction (will need @nowarn if this is OK)
                self._warn('Found no label for operand: %s %s' % (instruction.address, operation))
        elif label_warn and self.mode.do_ssubs and operand and int(operand) > 23755:
            # Warn if the operand is > 23755 (where code might be) but doesn't
            # refer to the address of an instruction (will need @nowarn if this
            # is OK)
            self._warn('Unreplaced operand: %s %s' % (instruction.address, operation))

    def _calculate_tap_routine_usage(self):
        tap_range = self.tap_info.get_tap_range()
        for t in tap_range:
            tap_address = self.get_word(self.tap_info.tap_address_table + t - self.tap_info.min_tap)
            tap_entry = self.get_entry(tap_address)
            if tap_entry:
                sep_index = tap_entry.description.index(':')
                self.tap_descs[t] = tap_entry.description[sep_index + 1:].strip()
            else:
                # Don't calculate tap routine usage if we don't have a full
                # snapshot
                return
            self.tap_addresses[t] = tap_address
        for t in tap_range:
            tap_address = self.tap_addresses[t]
            addr = tap_address
            while True:
                if addr >= 65535:
                    self._warn("Couldn't find end of command list %i" % t)
                    break
                routine_address = self.get_word(addr)
                routine = self.get_container(routine_address)
                routine.add_tap_referrer((t, tap_address))
                addr += 2
                num_params = self.tap_info.get_num_params(routine_address)
                params = self.snapshot[addr:addr + num_params]
                if params:
                    if self.tap_info.is_controller(routine_address):
                        subroutine_address = params[0] + 256 * params[1]
                        subroutine = self.get_container(subroutine_address)
                        subroutine.add_tap_referrer((t, tap_address))
                    addr += num_params
                if self.tap_info.is_terminal_command(routine_address, params):
                    break

class Mode:
    def __init__(self, level=0, verbose=False, warn=True, fix_mode=0):
        self.html = level == 0
        self.asm = level > 0
        self.verbose = verbose
        self.warn = warn
        self.started = level == 0
        self.include = self.started
        self.do_rfixes = fix_mode >= 3
        self.do_rsubs = level >= 3 or self.do_rfixes
        self.do_ssubs = level >= 2
        self.do_ofixes = fix_mode >= 1 or self.do_rsubs
        self.do_bfixes = fix_mode >= 2
        self.labels = []
        self.reset()

    def reset(self):
        self.label = None
        self.rlabel = None
        self.isub = None
        self.ofix = None
        self.bfix = None
        self.ssub = None
        self.rsub = None
        self.keep = False
        self.nowarn = False
        self.nolabel = False
        self.ignoreua = False
        self.ignoremrcua = False
        self.org = None

    def start(self):
        self.started = True
        self.include = True

    def apply_asm_attributes(self, instruction):
        if self.rlabel and self.do_rsubs:
            label = self.rlabel
        else:
            label = self.label
        if label:
            if label in self.labels:
                self._error('Duplicate label %s at %s' % (label, instruction.address))
            self.labels.append(label)
            instruction.asm_label = label

        if self.html:
            return
        sub = self.isub
        if self.bfix is not None and self.do_bfixes:
            sub = self.bfix
        elif self.ofix is not None and self.do_ofixes:
            sub = self.ofix
        elif self.rsub is not None and self.do_rsubs:
            sub = self.rsub
        elif self.ssub is not None and self.do_ssubs:
            sub = self.ssub
        if sub is not None:
            instruction.apply_sub(sub)

        instruction.warn = not self.nowarn
        instruction.nolabel = self.nolabel
        instruction.ignoreua = self.ignoreua
        instruction.ignoremrcua = self.ignoremrcua
        instruction.keep = self.keep
        instruction.org = self.org

        self.reset()

class Instruction:
    def __init__(self, ctl, address, operation):
        self.ctl = ctl
        self.address = address
        self.label = ctl + address
        self.operation = operation
        self.container = None
        self.reference = None
        self.comment = None
        self.referrers = []
        self.asm_label = None
        # If this instruction has no address, it was inserted between
        # @rsub+begin and @rsub+end; in that case, mark it as a subbed
        # instruction already
        self.sub = None if address.isdigit() else operation
        self.keep = False
        self.warn = True
        self.ignoreua = False
        self.ignoremrcua = False

    def set_comment(self, rowspan, text):
        self.comment = Comment(rowspan, text)

    def set_reference(self, routine_address, address):
        self.reference = Reference(routine_address, address)

    def add_referrer(self, routine):
        if routine not in self.referrers:
            self.referrers.append(routine)
        self.container.add_referrer(routine)

    def apply_sub(self, sub):
        self.sub = sub
        self.operation = sub

    def is_in_routine(self):
        return self.container.is_routine()

    def get_mid_routine_comment(self):
        return self.container.get_mid_routine_comment(self.label)

class Reference:
    def __init__(self, routine_address, address):
        self.routine_address = routine_address
        self.address = address

class Comment:
    def __init__(self, rowspan, text):
        self.rowspan = rowspan
        self.text = text

class SkoolEntry:
    def __init__(self, address, ctl=None, description=None, details=None):
        self.address = address
        self.ctl = ctl
        self.description = description
        self.details = details
        self.registers = []
        self.instructions = []
        self.mid_routine_comments = {}
        self.end_comment = None
        self.referrers = []
        self.size = None

    def is_routine(self):
        return False

    def is_data(self):
        return False

    def add_instruction(self, instruction):
        instruction.container = self
        self.instructions.append(instruction)

    def get_mid_routine_comment(self, address):
        return self.mid_routine_comments.get(address)

    def add_mid_routine_comment(self, address, text):
        self.mid_routine_comments[address] = text

    def add_referrer(self, routine):
        if routine not in self.referrers:
            self.referrers.append(routine)

class Routine(SkoolEntry):
    def __init__(self, address, ctl, description, details, registers):
        SkoolEntry.__init__(self, address, ctl, description, details)
        for name, contents in registers:
            self.registers.append(Register(name, contents))
        self.tap_referrers = []

    def is_routine(self):
        return True

    def add_tap_referrer(self, tap):
        if tap not in self.tap_referrers:
            self.tap_referrers.append(tap)

class Register:
    def __init__(self, name, contents):
        self.name = name
        self.contents = contents

class Data(SkoolEntry):
    def is_data(self):
        return True

class TableParser:
    TABLE_MARKER = '#TABLE'
    TABLE_END_MARKER = 'TABLE#'
    COLUMN_WRAP_MARKER = ':w'

    def parse_text(self, text, index):
        end_marker = self.__class__.TABLE_END_MARKER
        end = text.index(end_marker, index) + len(end_marker)
        table = self.parse_table(text[index:end])
        return end, table

    def parse_table(self, text):
        marker = self.__class__.TABLE_MARKER
        index = len(marker) if text.startswith(marker) else 0
        classes = []
        if text[index] == '(':
            end = text.index(')', index)
            classes = [c.strip() for c in text[index + 1:end].split(',')]
            index = end + 1
        table_class = classes[0] if classes else ''
        column_classes = classes[1:]
        wrap_columns = []
        for i, column_class in enumerate(column_classes):
            if column_class.endswith(self.__class__.COLUMN_WRAP_MARKER):
                column_classes[i] = column_class[:-len(self.__class__.COLUMN_WRAP_MARKER)]
                wrap_columns.append(i)
        table = Table(table_class, wrap_columns)

        prev_spans = {}
        while text.find('{', index) >= 0:
            row = []
            row_start = text.find('{ ', index)
            row_end = text.index(' }', row_start)
            col_index = 0
            for cell in text[row_start + 1:row_end + 1].split(' | '):
                prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                while prev_rowspan > 1:
                    prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
                    col_index += prev_colspan
                    prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                header, transparent = False, False
                rowspan, colspan = 1, 1
                cell_class = column_classes[col_index] if len(column_classes) > col_index else ''
                cell = cell.strip()
                if cell.startswith('='):
                    end = cell.find(' ')
                    if end < 0:
                        end = len(cell)
                    for span in cell[1:end].split(','):
                        if span[0] == 'c':
                            colspan = int(span[1:])
                        elif span[0] == 'r':
                            rowspan = int(span[1:])
                        elif span[0] == 'h':
                            header = True
                        elif span[0] == 't':
                            transparent = True
                    cell = cell[end:].lstrip()
                row.append(Cell(cell, transparent, colspan, rowspan, header, cell_class))
                prev_spans[col_index] = (rowspan, colspan)
                col_index += colspan

            # Deal with the case where the previous row contains one or more
            # cells at the end with rowspan > 1
            while col_index in prev_spans:
               prev_rowspan, prev_colspan = prev_spans[col_index]
               if prev_rowspan > 1:
                   prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
               col_index += prev_colspan

            table.add_row(row)
            index = row_end + 1

        return table

class Table:
    def __init__(self, table_class, wrap_columns):
        self.table_class = table_class
        self.wrap_columns = wrap_columns
        self.rows = []
        self.cells = []
        self.col_widths = None
        self.cell_padding = None
        self.num_cols = None

    def add_row(self, cells):
        self.rows.append(cells)
        self.cells.extend(cells)

    def prepare_cells(self):
        # For each cell, set row_index and col_index, and convert the contents
        # to a 1-item list. Also calculate num_cols and col_widths.
        prev_spans = {}
        for row_index, row in enumerate(self.rows):
            col_index = 0
            for cell in row:
                while True:
                    prev_rowspan, prev_colspan = prev_spans.get(col_index, (1, 1))
                    if prev_rowspan == 1:
                        break
                    prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
                    col_index += prev_colspan
                cell.row_index, cell.col_index = row_index, col_index
                cell.contents = [cell.contents]
                prev_spans[col_index] = (cell.rowspan, cell.colspan)
                col_index += cell.colspan

            # Deal with cells at the end of the previous row that have rowspan
            # greater than 1
            while col_index in prev_spans:
               prev_rowspan, prev_colspan = prev_spans[col_index]
               if prev_rowspan > 1:
                   prev_spans[col_index] = (prev_rowspan - 1, prev_colspan)
               col_index += prev_colspan

        self.num_cols = 1 + max([cell.col_index for cell in [row[-1] for row in self.rows]])
        self._calculate_col_widths()

    def get_header_rows(self):
        headers = []
        for i, row in enumerate(self.rows):
            if len([c for c in row if not (c.transparent or c.header)]) == 0:
                headers.append(i)
        return headers

    def get_cell_width(self, col_index, colspan):
        return sum(self.col_widths[col_index:col_index + colspan]) + self.cell_padding * (colspan - 1)

    def _calculate_col_widths(self):
        # Loop over the cells to collect their widths (considering only those
        # cells with colspan=1 on this pass)
        cell_widths = [[0] * self.num_cols for row in self.rows]
        for cell in [c for c in self.cells if c.colspan == 1]:
            cell_widths[cell.row_index][cell.col_index] = cell.get_width()

        # Scan each column to find the widest cell; the width of that cell will
        # be the width of the column
        self.col_widths = [max([row[i] for row in cell_widths]) for i in range(self.num_cols)]

        # Scan the cells again to make sure that those with colspan > 1 have
        # enough space
        for cell in [c for c in self.cells if c.colspan > 1]:
            col_index, colspan = cell.col_index, cell.colspan
            space_needed = cell.get_width() - self.get_cell_width(col_index, colspan)
            while space_needed > 0:
                for j in range(colspan):
                    if space_needed > 0:
                        self.col_widths[col_index + j] += 1
                        space_needed -= 1

    def reduce_width(self, max_table_width, min_col_width):
        # Reduce the width of the wrappable columns until either the table is
        # narrow enough or all the wrappable columns are the minimum width
        done = False
        while self.get_width() > max_table_width and not done:
            done = True
            for wrap_col in self.wrap_columns:
                if self.col_widths[wrap_col] > min_col_width:
                    self.col_widths[wrap_col] -= 1
                    done = False

        # Wrap the contents of the cells in the wrappable columns
        for cell in [c for c in self.cells if c.col_index in self.wrap_columns]:
            width = self.get_cell_width(cell.col_index, cell.colspan)
            cell.contents = textwrap.wrap(cell.contents[0], width) or ['']

        # The column widths need to be recalculated now
        self._calculate_col_widths()

    def get_width(self):
        padding = self.cell_padding * (self.num_cols + 1) - 2
        return padding + sum(self.col_widths)

class Cell:
    def __init__(self, contents, transparent, colspan, rowspan, header, cell_class):
        self.contents = contents
        self.colspan = colspan
        self.rowspan = rowspan
        self.header = header
        self.transparent = transparent
        self.cell_class = cell_class
        self.row_index = None
        self.col_index = None

    def get_width(self):
        return max([len(line) for line in self.contents])

class RefParser:
    def __init__(self):
        self.ref = []
        self.sections = {}

    def parse(self, reffile):
        infile = open(reffile, 'r')
        section_name = None
        section_lines = []
        for line in infile:
            if line.startswith('['):
                if section_name:
                    self.ref.append((section_name, section_lines))
                    self.sections[section_name] = section_lines
                section_name = line[1:line.index(']')]
                section_lines = []
            elif not line.startswith(';'):
                section_lines.append(line.rstrip())
        infile.close()
        self.ref.append((section_name, section_lines))
        self.sections[section_name] = section_lines

    def has_ref_items(self, section_type):
        prefix = section_type + ':'
        for section_name in self.sections:
            if section_name.startswith(prefix):
                return True
        return False

    def build_dictionary(self, section_name):
        dictionary = {}
        for line in self.sections.get(section_name, ()):
            if line:
                key, value = line.split('=')
                if key.isdigit():
                    key = int(key)
                dictionary[key] = value
        return dictionary

    def get_section(self, section_name):
        paragraphs = self._get_paragraphs(self.sections.get(section_name, ()), True)
        lines = ['\n'.join(p) for p in paragraphs]
        return '\n'.join(lines)

    def get_ref_items(self, section_type, keep_lines=False):
        prefix = section_type + ':'
        items = []
        for section_name, lines in self.ref:
            if section_name.startswith(prefix):
                elements = section_name.split(':', 2)
                paragraphs = self._get_paragraphs(lines, keep_lines)
                if len(elements) < 3:
                    items.append((elements[1], paragraphs))
                else:
                    items.append((elements[1], elements[2], paragraphs))
        return items

    def get_tap_info(self, address_table):
        num_params = []
        controllers = []
        terminal_commands = []
        for line in self.sections['TapInfo']:
            if not line:
                continue
            key, value = line.split('=')
            if key == 'MaxTap':
                max_tap = int(value)
            elif key == 'NumParams':
                num_params.append([int(i) for i in value.split(',')])
            elif key == 'Controller':
                controllers.append(int(value))
            elif key == 'TerminalCommand':
                terminal_commands.append([int(i) for i in value.split(',')])
        tap_info = TapInfo(address_table, max_tap)
        for routine, num_p in num_params:
            tap_info.set_num_params(routine, num_p)
        for controller in controllers:
            tap_info.add_controller(controller)
        for args in terminal_commands:
            if len(args) == 1:
                tap_info.add_terminal_command(args[0])
            else:
                tap_info.add_terminal_command(args[0], [args[1]])
        return tap_info

    def _get_paragraphs(self, lines, keep_lines):
        paragraphs = []
        p = []
        for line in self._expand_includes(lines):
            if line:
                p.append(line)
            elif p:
                paragraphs.append(p if keep_lines else '\n'.join(p))
                p = []
        if p:
            paragraphs.append(p if keep_lines else '\n'.join(p))
        return paragraphs

    def _expand_includes(self, lines):
        new_lines = []
        for line in lines:
            if line.startswith('::'):
                inc_section = self.sections.get(line[2:], ())
                if inc_section:
                    new_lines.extend(self._expand_includes(inc_section))
            else:
                new_lines.append(line)
        return new_lines

class TapInfo:
    def __init__(self, tap_address_table, max_tap):
        self.tap_address_table = tap_address_table
        self.min_tap = tap_address_table % 256
        self.max_tap = max_tap
        self.tap_routine_params = {}
        self.tap_controllers = []
        self.terminal_commands = {}

    def get_tap_range(self):
        return range(self.min_tap, self.max_tap + 1, 2)

    def get_num_params(self, address):
        return self.tap_routine_params.get(address) or 0

    def is_terminal_command(self, address, params):
        return address in self.terminal_commands and self.terminal_commands[address] == params

    def is_controller(self, address):
        return address in self.tap_controllers

    def set_num_params(self, address, num_params):
        self.tap_routine_params[address] = num_params

    def add_controller(self, address):
        self.tap_controllers.append(address)

    def add_terminal_command(self, address, params=None):
        self.terminal_commands[address] = params or []
